@@ -3,7 +3,7 @@
 """BBRF Client
 
 Usage:
-  bbrf ( new | use | disable | enable ) <program> [ -t key:value ]...
+  bbrf ( new | use | disable | enable | rm ) <program> [ -t key:value ]...
   bbrf programs [ --show-disabled --show-empty-scope ]
   bbrf programs where <tag_name> is [ before | after ] <value> [ --show-disabled --show-empty-scope ]
   bbrf program ( active | update ( <program>... | - ) -t key:value... [--append-tags])
@@ -16,7 +16,7 @@ Usage:
   bbrf scope ( in | out ) [ (--wildcard [--top] ) ] [ ( -p <program> ) | ( --all [--show-disabled] ) ]
   bbrf scope filter ( in | out ) [ (--wildcard [--top] ) ] [ ( -p <program> ) | ( --all [--show-disabled] ) ]
   bbrf ( inscope | outscope ) ( add | remove ) ( - | <element>... ) [ -p <program> ]
-  bbrf urls [ -d <hostname> | ( -p <program> | ( --all [--show-disabled] ) ) ] [--with-query]
+  bbrf urls [ -d <hostname> | ( -p <program> | ( --all [--show-disabled] ) ) ] [ --with-query | --root ]
   bbrf urls where <tag_name> is [ before | after ] <value> [ -p <program> | ( --all [--show-disabled] ) ]
   bbrf url add ( - | <url>... ) [ -d <hostname> -s <source> -p <program> --show-new ( -t key:value... [--append-tags] ) ]
   bbrf url remove ( - | <url>... )
@@ -45,6 +45,7 @@ Options:
   -A, --all            Specify to get information across all programs. Incompatible with the -p flag
   -a, --append-tags    If a tag with the same name already exists on the document, make an array and append the new value(s)
   -q, --with-query     When listing URLs, show all URLs including queries
+  -r, --root           When listing URLs, only list the roots of the URLs
   -w, --show-disabled  Combine with the flag --all/-A to include documents of disabled programs too
 """
 
@@ -61,7 +62,7 @@ CONFIG_FILE = '~/.bbrf/config.json'
 REGEX_DOMAIN = re.compile('^(?:[a-z0-9_](?:[a-z0-9-_]{0,61}[a-z0-9])?\\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$')
 # regex to match IP addresses and CIDR ranges - thanks https://www.regextester.com/93987
 REGEX_IP = re.compile('^([0-9]{1,3}\\.){3}[0-9]{1,3}(/([0-9]|[1-2][0-9]|3[0-2]))?$')
-VERSION = '1.1.9'
+VERSION = '1.1.11'
 
 class BBRFClient:
     config = {}
@@ -114,7 +115,7 @@ class BBRFClient:
         self.api.create_new_program(self.get_program(), tags=self.arguments['-t'])
 
     def list_programs(self, show_disabled, show_empty_scope):
-        return self.api.get_programs(show_disabled, show_empty_scope=show_empty_scope)
+        return self.api.get_programs(show_disabled=show_disabled, show_empty_scope=show_empty_scope)
     
     # updating programs this way only supports tags for now
     def update_programs(self, programs):
@@ -257,7 +258,7 @@ class BBRFClient:
             domain = domain.lower()
             
             if ':' in domain:
-                domain, ips = domain.split(':')
+                domain, ips = domain.split(':',1)
                 ips = ips.split(',')
                 
                 for ip in ips:
@@ -323,8 +324,7 @@ class BBRFClient:
                 return ["[NEW] "+(x if not x.startswith('._') else x[1:]) for x in success if x]
     
     '''
-    This is now balanced over 100 concurrent threads in order to drastically
-    improve the throughput of these requests.
+    Use the bulk update function to mass remove domains.
     '''
     def remove_domains(self, domains):
         
@@ -333,6 +333,22 @@ class BBRFClient:
         
         if self.arguments['--show-new'] and removed:
             return ["[DELETED] "+(x if not x.startswith('._') else x[1:]) for x in removed if x]
+        
+    '''
+    Remove all documents related to a program.
+    '''
+    def remove_program(self, program):
+        all_docs = self.api.get_all_program_documents(program)
+        remove_domains = {docid: {'_deleted': True} for (docid, doctype) in all_docs if doctype == 'domain' }
+        remove_ips = {docid: {'_deleted': True} for (docid, doctype) in all_docs if doctype == 'ip' }
+        remove_urls = {docid: {'_deleted': True} for (docid, doctype) in all_docs if doctype == 'url' }
+        remove_services = {docid: {'_deleted': True} for (docid, doctype) in all_docs if doctype == 'service' }
+        
+        self.api.update_documents('domain', remove_domains)
+        self.api.update_documents('ip', remove_ips)
+        self.api.update_documents('url', remove_urls)
+        self.api.update_documents('service', remove_services)
+        self.api.update_documents('program', {program: {'_deleted': True}})
 
     '''
     Update properties of a domain
@@ -506,8 +522,9 @@ class BBRFClient:
                 
             # parse properties of the URL
             u = urlparse(url)
+            scheme = u.scheme
             port = u.port
-            query = u.query if not u.query == "" else None
+            query = u.query if u.query else None
             
             # Usually we parse the hostname from the URL,
             # but we don't need to if a hostname is explicitly set
@@ -521,13 +538,13 @@ class BBRFClient:
             # We can't process relative URLS without understanding,
             # so need to skip those for now. Better ideas wecome!
             if not hostname:
-                print("Hostname could not be parsed, skipping "+url)
+                self.debug("Hostname could not be parsed, skipping "+url)
                 continue
             
             # If the provided hostname in -d does not match the parsed hostname,
             # we won't add it to avoid polluting the dataset
             if u.hostname and not u.hostname == hostname:
-                print("Provided hostname "+hostname+" did not match parsed hostname "+u.hostname+", skipping...")
+                self.debug("Provided hostname "+hostname+" did not match parsed hostname "+u.hostname+", skipping...")
                 continue
                 
             # If the provided URL is relative, we need to rewrite the URL
@@ -535,32 +552,41 @@ class BBRFClient:
             if not u.netloc:
                 url = 'http://' + hostname + url
                 port = 80
+                scheme = 'http'
             if url.startswith('//'):
                 url = 'http:' + url
                 port = 80
+                scheme = 'https'
             if '?' in url:
                 url = url.split('?')[0]
                 
             # It must match the format of a domain name or an IP address
             if not REGEX_DOMAIN.match(hostname) and not REGEX_IP.match(hostname):
-                print("Illegal hostname:",hostname)
+                self.debug("Illegal hostname:",hostname)
                 continue
             # It may not be explicitly outscoped
             if not self.get_program() == '@INFER' and self.matches_scope(hostname, outscope):
-                print("skipping outscoped hostname:",hostname)
+                self.debug("skipping outscoped hostname:",hostname)
                 continue
             # It must match the in scope
             if not self.get_program() == '@INFER' and not self.matches_scope(hostname, inscope):
-                print("skipping not inscope hostname:",hostname)
+                self.debug("skipping not inscope hostname:",hostname)
                 continue
             
             if not port:
-                if u.scheme == 'http':
+                if scheme == 'http':
                     port = 80
-                if u.scheme == 'https':
+                if scheme == 'https':
                     port = 443
             
-            if url in add_urls and query not in add_urls[url]['query']:
+            # remove explicit ports 80 and 443 to avoid duplicate values
+            # when they are also added without the ports
+            if port == 80 and ':80/' in url:
+                url = url.replace(':80/', '/', 1)
+            elif port == 443 and ':443/' in url:
+                url = url.replace(':443/', '/', 1)
+            
+            if url in add_urls and query and query not in add_urls[url]['query']:
                 add_urls[url]['query'].append(query)
             else:
                 if len(parts) == 1: # only a url
@@ -568,7 +594,9 @@ class BBRFClient:
                         add_urls[url] = {
                             "hostname": hostname,
                             "port": int(port),
-                            "query": [query] if query else []
+                            "query": [query] if query else [],
+                            "scheme": scheme,
+                            "path": u.path
                         }
 
                 elif len(parts) == 3: # url, status code and content length
@@ -577,7 +605,9 @@ class BBRFClient:
                         "port": int(port),
                         "status": int(parts[1]),
                         "content_length": int(parts[2]),
-                        "query": [query] if query else []
+                        "query": [query] if query else [],
+                        "scheme": scheme,
+                        "path": u.path
                     }
         
         success, failed = self.api.add_documents('url', add_urls, self.get_program(), source=self.arguments['-s'], tags=self.arguments['-t'])
@@ -789,13 +819,13 @@ class BBRFClient:
     
     def list_urls(self, by, key = False):
         if by == "hostname":
-            return self.api.get_urls_by_hostname(key, with_query=self.arguments['--with-query'])
+            return self.api.get_urls_by_hostname(key, with_query=self.arguments['--with-query'], root_only=self.arguments['--root'])
         elif by == "program":
-            return self.api.get_urls_by_program(key, with_query=self.arguments['--with-query'])
+            return self.api.get_urls_by_program(key, with_query=self.arguments['--with-query'], root_only=self.arguments['--root'])
         elif by == "all":
-            return self.api.get_urls_by_program(with_query=self.arguments['--with-query'], show_disabled=self.arguments['--show-disabled']) # An empty key will return all results
+            return self.api.get_urls_by_program(with_query=self.arguments['--with-query'], root_only=self.arguments['--root'], show_disabled=self.arguments['--show-disabled']) # An empty key will return all results
         else:
-            return self.api.get_urls_by_program(self.get_program(), with_query=self.arguments['--with-query'])
+            return self.api.get_urls_by_program(self.get_program(), with_query=self.arguments['--with-query'], root_only=self.arguments['--root'])
         
     def list_services(self, list_all = False):
         if list_all:
@@ -901,6 +931,9 @@ class BBRFClient:
             
         if self.arguments['enable']:
             self.enable_program(self.arguments['<program>'][0])
+            
+        if self.arguments['rm']:
+            self.remove_program(self.arguments['<program>'][0])
 
         if self.arguments['programs']:
             if self.arguments['where']:
