@@ -81,7 +81,7 @@ class BBRFApi:
     '''
     Get a list of all urls, filtered by program or hostname if provided.
     '''
-    def get_urls_by_hostname(self, hostname=None, with_query=False):
+    def get_urls_by_hostname(self, hostname=None, with_query=False, root_only=False):
         if hostname:
             r = self.requests_session.get(self.BBRF_API+'/_design/bbrf/_view/urls_by_hostname?reduce=false&key="'+hostname+'"', headers={"Authorization": self.auth})
         else:
@@ -89,12 +89,12 @@ class BBRFApi:
         if 'error' in r.json():
             raise Exception('BBRF server error: '+r.json()['error'])
         # print all url, status, content_length if status and content length are set    
-        return self.process_urls(r.json()['rows'], with_query)
+        return self.process_urls(r.json()['rows'], with_query, root_only)
     
-    def get_urls_by_program(self, program=None, with_query=False, show_disabled=False):
+    def get_urls_by_program(self, program=None, with_query=False, root_only=False, show_disabled=False):
         if program:
             r = self.requests_session.get(self.BBRF_API+'/_design/bbrf/_view/urls_by_program?reduce=false&key="'+requests.utils.quote(program, safe='')+'"', headers={"Authorization": self.auth})
-            return self.process_urls(r.json()['rows'], with_query)
+            return self.process_urls(r.json()['rows'], with_query, root_only)
         else:
             r = self.requests_session.get(self.BBRF_API+'/_design/bbrf/_view/urls_by_program?reduce=false', headers={"Authorization": self.auth})
             if show_disabled:
@@ -102,12 +102,16 @@ class BBRFApi:
                 return self.process_urls(r.json()['rows'], with_query)
             else:
                 active_programs = self.get_programs(show_disabled=show_disabled)
-                return self.process_urls([x for x in r.json()['rows'] if x['key'] in active_programs], with_query)        
+                return self.process_urls([x for x in r.json()['rows'] if x['key'] in active_programs], with_query, root_only)        
         
-    def process_urls(self, urls, with_query=False):
+    def process_urls(self, urls, with_query=False, root_only=False):
         if not with_query:
+            if not root_only:
             # print all url, status, content_length if status and content length are set
-            return [" ".join([str(x) for x in r['value'][:3] if x]) for r in urls]
+                return [" ".join([str(x) for x in r['value'][:3] if x]) for r in urls]
+            else:
+            # print only roots or urls:
+                return list(set(['/'.join(r['value'][0].split('/',3)[:3]) for r in urls]))
         else:
             # get a list of URLs without any queries
             no_query = [" ".join([str(x) for x in r['value'][:3] if x]) for r in urls if len(r['value'][3]) == 0]
@@ -251,6 +255,15 @@ class BBRFApi:
         r = self.requests_session.put(self.BBRF_API+'/'+requests.utils.quote(program_name, safe='')+'?rev='+program['_rev'], json.dumps(program), headers={"Authorization": self.auth})
         if 'error' in r.json():
             raise Exception('BBRF server error: '+r.json()['error'])
+        
+    '''
+    Return all document identifiers of documents belonging to a program.
+    '''
+    def get_all_program_documents(self, program_name):
+        r = self.requests_session.get(self.BBRF_API+'/_design/bbrf/_view/program_all_documents?key="'+requests.utils.quote(program_name, safe='')+'"', headers={"Authorization": self.auth})
+        if 'error' in r.json():
+            raise Exception('BBRF server error: '+r.json()['error'])
+        return [ (r['id'], r['value']) for r in r.json()['rows'] ]
             
     '''
     Get a list of all agents.
@@ -322,7 +335,6 @@ class BBRFApi:
                     elif doctype == 'url':
                         from . import BBRFClient
                         infered_program_name = [p for p in program_scopes.keys() if BBRFClient.matches_scope(identifiers[docid]['hostname'], program_scopes[p]['in']) and not BBRFClient.matches_scope(identifiers[docid]['hostname'], program_scopes[p]['out'])]
-                        print(infered_program_name)
                         if len(infered_program_name) > 0:
                             infered_program_name = infered_program_name[0]
 
@@ -819,9 +831,11 @@ class BBRFApi:
         # filter results based on the doctype and program name
         if doctype:
             results = [x for x in results if x['value'][0] == doctype]
+        # restrict further if needs to match a specific program
         if not doctype == 'program' and program:
             results = [x for x in results if x['value'][2] == program]
-        if not show_disabled and not doctype == 'program':
+        # remove disabled programs unles the --show-disabled flag is set
+        if not show_disabled and doctype == 'program':
             active_programs = self.get_programs(show_disabled=show_disabled, show_empty_scope=show_empty_scope)
             results = [x for x in results if x['value'][2] in active_programs]
         
